from __future__ import annotations

import datetime
from dataclasses import dataclass
from decimal import Decimal


@dataclass
class Transaction:
    """
    Representation of a Transaction.
    """

    date: datetime.datetime
    description: str
    amount: Decimal
    payee: str
    account: str

    def to_ledger(self, config):
        return f"""{self.date.strftime("%Y/%m/%d")}    {self.description}
    {self.account}    {config.format_amount(self.amount)}
    {self.payee}
"""



from __future__ import annotations

import sys

import _csv

from ledger_importer.config import Config
from ledger_importer.transaction import Transaction


class TransactionsHandler:
    config: Config

    def __init__(self, config: Config) -> None:
        self.config = config

    def row_to_transaction(self, row: tuple) -> Transaction:
        return Transaction(
            date=self.config.parse_date(row),
            description=self.config.parse_description(row),
            amount=self.config.parse_amount(row),
            payee=self.config.parse_payee(row),
            account=self.config.parse_account(row),
        )

    def merge_transactions(self, transactions: list[Transaction]) -> list[Transaction]:
        """
        Take a list of transactions already sorted by date (oldest -> newest).
        Update the payees of the transactions when they seem to match.

        The matching is tried only on positive transactions, in this case
        it will look for older matching transactions.
        """
        merged_transactions: list[Transaction] = []

        for transaction in transactions:
            if transaction.amount > 0:
                for matching_transaction in transactions:
                    if matching_transaction is transaction:
                        # Nothing was found up to the current transaction
                        break

                    if self.config.transactions_match(transaction, matching_transaction):
                        transaction.payee = matching_transaction.account
                        merged_transactions.append(matching_transaction)
                        break

        return [transaction for transaction in transactions if transaction not in merged_transactions]

    def parse_transactions(self, csv_reader: _csv._reader) -> list[Transaction]:
        """
        Parse transactions from the csv reader and sort them chronologically.
        """
        transactions: list[Transaction] = []

        for _ in range(self.config.skip_lines):
            next(csv_reader)

        for row in csv_reader:
            transaction = self.row_to_transaction(tuple(row))
            transactions.append(transaction)

        return sorted(transactions, key=lambda transaction: transaction.date)

    def confirm_transactions(self, transactions: list[Transaction]) -> list[Transaction]:
        """
        Manually confirm the transactions using the cli.

        Returns only confirmed transactions.
        """
        confirmed_transactions: list[Transaction] = []

        for transaction in transactions:
            print(
                f"""
| {"Account".center(max(len(transaction.account), len("Account")))} | {"Date".center(10)} | {"Amount".center(max(len(self.config.format_amount(transaction.amount)), len("Amount")))} | {"Description".center(max(len(transaction.description), len("Description")))} |
| {transaction.account} | {transaction.date.strftime("%Y/%m/%d")} | {self.config.format_amount(transaction.amount)} | {transaction.description} |
    """,
                file=sys.stderr,
            )

            if transaction.amount > 0:
                print(
                    f"""Which account provided this income? ([{transaction.payee}]/[q]uit/[s]kip) """,
                    file=sys.stderr,
                )
            else:
                print(
                    f"""To which account did this money go? ([{transaction.payee}]/[q]uit/[s]kip) """,
                    file=sys.stderr,
                )

            answer = input()
            if answer == "q":
                break
            elif answer == "s":
                continue

            if answer:
                transaction = Transaction(
                    date=transaction.date,
                    description=transaction.description,
                    amount=transaction.amount,
                    payee=answer,
                    account=transaction.account,
                )
            confirmed_transactions.append(transaction)

        return confirmed_transactions



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ledger_importer']

package_data = \
{'': ['*']}

install_requires = \
['typer>=0.4.1,<0.5.0']

setup_kwargs = {
    'name': 'ledger-importer',
    'version': '0.2.0',
    'description': 'ledger_importer is a csv-to-ledger importer that can be configured in Python.',
    'long_description': '# ledger_importer [![CircleCI](https://circleci.com/gh/volnt/ledger_importer.svg?style=shield&circle-token=afb73aed03518c8658de39f5d61ec3bfdf50d57c)](https://app.circleci.com/settings/project/github/volnt/ledger_importer)\n\nledger_importer is a csv-to-[ledger](https://www.ledger-cli.org/3.0/doc/ledger3.html) importer that can be configured in Python.\n\nThe key features are:\n\n* **Customization**: Designed to fit your specific needs perfectly.\n* **Auto-completion**: The confirmation step is auto-completed.\n* **Integration**: Easy to integrate with your pipeline.\n\n## Installation\n\n```sh\n$ pip install ledger-importer\n```\n\n## Configure\n\nledger_importer works by using the configuration file as the entrypoint. The `ledger_importer.runner` function is the function that should be called when you want to run the program.\n\nThe `runner` function takes a Config as the first argument.\n\nA Config instance can be created by creating a new class that inherits `ledger_importer.runner`. This new class must implement the following methods:\n\n* `parse_date(self, fields: tuple) -> datetime.datetime`\n* `parse_description(self, fields: tuple) -> str`\n* `parse_amount(self, fields: tuple) -> Decimal`\n* `format_amount(self, amount: Decimal) -> str`\n* `parse_payee(self, fields: tuple) -> str`\n* `parse_account(self, fields: tuple) -> str`\n\nThe argument `fields: tuple` will be the csv row, with each column as an element of the tuple.\n\n\nExample configuration file:\n\n```py\n#!/usr/bin/env python\nfrom __future__ import annotations\n\nimport datetime\nimport re\nfrom decimal import Decimal\n\nfrom ledger_importer import Config, runner\n\n# Custom ledger importer configuration\nclass LedgerImporterConfig(Config):\n    # Define the number of lines that needs to be skipped at the beginning of the file.\n    # This is usefull if the csv has a line with the column names for example.\n    skip_lines: int = 1\n\n    # The argument `fields` given in all parse_* methods contains a whole csv row in a tuple\n    # Each element of the tuple is a string representation of the column\n\n    def parse_date(self, fields: tuple) -> datetime.datetime:\n        return datetime.datetime.strptime(fields[0], "%m-%d-%Y")\n\n    def parse_description(self, fields: tuple) -> str:\n        return fields[2]\n\n    def parse_amount(self, fields: tuple) -> Decimal:\n        return Decimal(re.sub("[€$, ]", "", fields[3]))\n\n    def format_amount(self, amount: Decimal) -> str:\n        return f"${amount}"\n\n    def parse_payee(self, fields: tuple) -> str:\n        if self.parse_amount(fields) > 0:\n            return "Income"\n\n        return "Expenses"\n\n    def parse_account(self, fields: tuple) -> str:\n        return "Assets:Checking"\n\n\n# The next lines are required to run ledger_importer\n# when the config file is executed.\nif __name__ == "__main__":\n    runner(LedgerImporterConfig())```\n\n## Run\n\nTo run leger_importer, run the configuration module:\n\n```sh\n$ python my_importer.py bank-statement.csv --journal-path journal.ledger\n\n|        Account         |    Date    |  Amount  |     Description     |\n| Assets:Account:Florent | 2021/07/29 | 1234.56€ | VIR LOLE FOOB A.R.L |\n\nWhich account provided this income? ([Income:Salary]/[q]uit/[s]kip)\n\n\n|        Account         |    Date    |  Amount |         Description          |\n| Assets:Account:Florent | 2021/08/02 | -11.77€ | CARD  27/07/21 SWILE XX*XXXX |\n\nTo which account did this money go? ([Expenses:Restaurant]/[q]uit/[s]kip)\n\n\n|        Account         |    Date    |   Amount  |               Description               |\n| Assets:Account:Florent | 2021/08/03 |  -784.00€ | VIR Save some € Mr.      Florent        |\n\nTo which account did this money go? ([Expenses]/[q]uit/[s]kip)\nAssets:Savings\n\n\n|        Account         |    Date    |  Amount |          Description          |\n| Assets:Account:Florent | 2021/08/03 | -58.63€ | CARD  08/03/21 PAYPAL XX*XXXX |\n\nTo which account did this money go? ([Expenses:Shopping]/[q]uit/[s]kip)\nq\n```\n\n## Usage\n\n```sh\n$ python my_importer.py --help\nUsage: my_importer.py [OPTIONS] CSV_PATH\n\nArguments:\n  CSV_PATH  Path to the bank statement to import.  [required]\n\nOptions:\n  --journal-path PATH             Path a ledger journal to write & learn\n                                  accounts from.\n  --install-completion [bash|zsh|fish|powershell|pwsh]\n                                  Install completion for the specified shell.\n  --show-completion [bash|zsh|fish|powershell|pwsh]\n                                  Show completion for the specified shell, to\n                                  copy it or customize the installation.\n  --help                          Show this message and exit.\n```\n',
    'author': 'Florent Espanet',
    'author_email': 'florent.esp@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/volnt/ledger_importer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import datetime
from abc import ABC
from abc import abstractmethod
from decimal import Decimal

from ledger_importer.transaction import Transaction


class Config(ABC):
    skip_lines: int = 1
    csv_delimiter: str = ","
    default_account: str = "Assets:Checking"

    @abstractmethod
    def parse_date(self, fields: tuple) -> datetime.datetime:
        pass

    @abstractmethod
    def parse_description(self, fields: tuple) -> str:
        pass

    @abstractmethod
    def parse_amount(self, fields: tuple) -> Decimal:
        pass

    @abstractmethod
    def format_amount(self, amount: Decimal) -> str:
        pass

    @abstractmethod
    def parse_payee(self, fields: tuple) -> str:
        pass

    @abstractmethod
    def parse_account(self, fields: tuple) -> str:
        pass

    def transactions_match(self, transaction1: Transaction, transaction2: Transaction) -> bool:
        """
        Matching transactions will get merged in a single transaction.

        When they are merged, the payee of transaction1 will be updated to the account of transaction2. This overrides the parse_payee return value.

        Default behavior is to merge transactions with opposite amounts and different accounts.
        """
        return transaction1.amount == -transaction2.amount and transaction1.account != transaction2.account
